import java.util.*;
import java.io.*;

// https://codejam.withgoogle.com/codejam/resources/quickstart-guide#io-tutorial
// 
// MY_PROGRAM < small_input.txt > small_output.txt

public class DiceMazes {

	/** Main method: parse input file, build mazes, solve mazes. */
	public static void main(String[] args) throws FileNotFoundException {
		//Scanner scanSizes = new Scanner(new BufferedReader(new InputStreamReader(System.in)));

		// Create an input file scanner
		String filename = "in.txt";
		Scanner scan = new Scanner(new BufferedReader(new FileReader(new File(filename))));

		// Three step process:
		// 1. process inputs (parse input data for different Maze objects) 
		// 2. build mazes (use input data to construct State Diagrams
		// 3. solve mazes (using recursive backtracking)
		List<Maze> mazes = processInput(scan);
		buildMazes(mazes);
		solveMazes(mazes);
	}

	/** Process input file: parse input file data corresponding to 
	  * different mazes and different maze data. */
	public static List<Maze> processInput(Scanner in) {

		List<Maze> mazes = new ArrayList<Maze>();

		// Probably safe to assume the input file has SOMETHING in it, but just in case...
		while(in.hasNextLine()){ 

			// Start with the puzzle name
			String mazeName = in.nextLine();
			Maze m;
			if(mazeName.equals("END")) { 
				break;
			}
			m = new Maze(mazeName);
			
			// First line is start/finish node
			m.addString(in.nextLine());

			// Remaining lines specify state diagram
			String nextLine = in.nextLine();
			while(!nextLine.equals("0")) {
				m.addString(nextLine);
				nextLine = in.nextLine();
			}

			// Add this maze to the list
			mazes.add(m);
		}

		return mazes;
	}

	/** Build mazes: use input file to construct State Diagram graphs. */
	public static void buildMazes(List<Maze> mazes){
		for(Maze m : mazes) { 
			m.build();
		}
	}

	/** Solve mazes: use recursive backtracking to find shortest paths on State Diagram graphs. */
	public static void solveMazes(List<Maze> mazes){
	}

}

class Maze {
	String mazeName;	// maze name
	List<String> lines; // maze inputs
	Dimension dim;		// maze dimensions

	Node head, tail;

	public Maze(String mazeName) { 
		this.mazeName = mazeName;
		this.lines = new ArrayList<String>();
	}

	/** Add a String from the input file, so the Maze can access it later. */
	public void addString(String s){ 
		lines.add(s);
	}

	/** Build the Maze from the input file data. */
	public void build() {
		// Build a maze:
		// 1. get maximum dimensions of maze
		// 2. create array to hold nodes, and build nodes 
		// 3. link the nodes

		// Get maximum dimensions
		Dimension d = getMaxDimensions();

		//// Create the nodes
		//createNodes(d);

		// Create/link the Nodes
		linkNodes();
	}

	/** Link the nodes that compose the State Diagram graph. */
	public void linkNodes() { 

		Map<String,Node> nodesMap = new HashMap<String,Node>();

		// Iterate through all lines and extract row/column indexes.
		// Then use the letter combinations that follow
		// to determine how the graph is connected.
		Scanner lineScan; 
		int i=0;
		for(String thisLine : this.lines) {

			System.out.println("------------------------------------------");
			System.out.println("Line: \"" + thisLine + "\"");
			System.out.println("------------------------------------------");

			// First line is different 
			if(i==0) { 

				// First line contains start and end locations.
				String[] tokens = thisLine.split(" ");



				// Start (Head) Node:
				//
				// The start node logic is a bit complicated.
				// We are given the starting intersection's row and column, 
				// and the direction in which things are moving.
				//
				// However, we are not representing these intersections
				// with Nodes, but rather are representing the hallways 
				// connecting the intersections (one Node = one direction).
				// 
				// We label the Node we will construct with the 
				// name of the hallway and the initial direction 
				// that we're moving in.
				//
				// Then we use that label to insert the Node into a map.
				//
				// The Node also needs to know what intersections it connects
				// from and to, so that it can assemble a pathway through the 
				// intersections for the final solution.

				// Starting intersection coordinates/direction:
				int startRow = Integer.parseInt(tokens[0]);
				int startCol = Integer.parseInt(tokens[1]);
				String startDir = tokens[2];

				// Assemble names of maze intersections
				// this node connects from and to.
				String headFrom = "("+startRow+","+startCol+")";
				String headTo = "";
				if(startDir.equals("N")) {
					headTo = "("+ (startRow-1) +","+startCol+")";

				} else if(startDir.equals("S")) { 
					headTo = "("+(startRow+1)+","+startCol+")";

				} else if(startDir.equals("E")) { 
					headTo = "("+(startRow)+","+(startCol+1)+")";

				} else if(startDir.equals("W")) { 
					headTo = "("+(startRow)+","+(startCol-1)+")";

				}

				// Now assemble the label for this head Node,
				// which consists of the intersection coordinates it starts from,
				// plus the direction in which we're moving.
				// 
				// This is used to look up the Node in a Map.
				String headLabel = headFrom+" "+startDir;

				// Finally, we construct the head Node: 
				// pass it the intersections it connects from and to,
				// and use its label when inserting it into the Map. 
				Node headNode = new Node(headLabel, headFrom, headTo);
				headNode.setHead();
				nodesMap.put(headLabel, headNode);



				// End (Tail) Node:
				//
				// The end node logic is also a bit complicated, for different reasons.
				// We are given the ending intersection's row and column,
				// but not the directions that can be taken to reach it.
				//
				// We label the Node we will construct with the name of the hallway,
				// and the direction. This direction can only be obtained, 
				// at this point, if we make assumptions about how the finish node
				// can be reached. If we assume there is only one way to reach the finish node,
				// we can just say that it is the opposite direction to the initial start direction.
				//
				// This is justifiable based on the two mazes that are given as examples,
				// both of which have the start and end nodes on the same row, 
				// and both of which provide only one final direction to travel in 
				// to reach the end node.

				// Ending intersection coordinates/direction: 
				int endRow = Integer.parseInt(tokens[3]);
				int endCol = Integer.parseInt(tokens[4]);

				// We have no way of telling the ending direction.
				//
				// We can assume that the ending direction
				// is always the opposite of the starting direction,
				// which would be a safe assumption based on the two 
				// examples that were given.
				String endDir = "";
				if(endRow==startRow) { 
					if(startDir.equals("N")) {
						endDir="S";
					} else if(startDir.equals("S")) {
						endDir="N";
					}

				} else if(endCol==startCol) { 
					if(startDir.equals("E")) {
						endDir="W";
					} else if(startDir.equals("W")) {
						endDir="E";
					}

				} else {
					throw new RuntimeException("ERROR: malformed input file. Start/end nodes should be on same row or same column.");
				}

				// Assemble label for this Node.
				// Because the end node is the destination intersection of the move,
				// and the label should use the source intersection of the move,
				// this looks a little different from above.
				String tailLabel;

				String tailTo = "("+endRow+","+endCol+")";
				String tailFrom = "";
				// Note: not typos - should be opposite of above.
				if(endDir.equals("N")) {
					tailFrom = "("+ (endRow+1) +","+endCol+")";

				} else if(endDir.equals("S")) { 
					tailFrom = "("+(endRow-1)+","+endCol+")";

				} else if(endDir.equals("E")) { 
					tailFrom = "("+(endRow)+","+(endCol-1)+")";

				} else if(endDir.equals("W")) { 
					tailFrom = "("+(endRow)+","+(endCol+1)+")";

				}

				// Now assemble the label for this tail Node:
				// starting intersection, plus direction.
				//
				// This is used to look up the Node in a Map.
				tailLabel = tailFrom+" "+endDir;

				// Finally, construct the tail Node: 
				Node tailNode = new Node(tailLabel, tailFrom, tailTo);
				tailNode.setTail();
				nodesMap.put(tailLabel, tailNode);



			} else {

				// Each remaining line of the input file consists of a line 
				// containing coordinates for a given intersection, 
				// followed by chunks of letters, 
				// each specifying the direction of approach to the intersection,
				// plus a list of one or more valid turns. 
				String[] tokens = thisLine.split(" ");
				int thisRow = Integer.parseInt(tokens[0]);
				int thisCol = Integer.parseInt(tokens[1]);

				// What remains are chunks of letters that specify
				// the valid directions of approach for this node.
				//
				// We will loop over each line and each chunk of letters:
				// create the nodes as necessary, and link them together.
				for(int j=2; j<tokens.length-1; j++) { 

					// Grab this particular chunk and turn it into characters.
					String chunk = tokens[j];

					// The direction of approach to the intersection
					// located at (thisRow, thisCol)
					String approachDir = chunk.substring(0,1);



					// We need to assemble the two labels 
					// of the two nodes we are connecting:
					//
					// sourceLabel
					// destinationLabel



					// SourceLabel: 
					//
					// We are potentially constructing this node,
					// so we need to specify what intersections it connects 
					// from/to. That's what thisTo and thisFrom are for.
					String sourceTo = "("+thisRow+","+thisCol+")";
					String sourceFrom = "";
					if(approachDir.equals("N")) { 
						sourceFrom = "(" + (thisRow+1) + "," + thisCol + ")";

					} else if(approachDir.equals("S")) { 
						sourceFrom = "("+(thisRow-1)+","+thisCol+")";

					} else if(approachDir.equals("E")) { 
						sourceFrom = "("+(thisRow)+","+(thisCol-1)+")";

					} else if(approachDir.equals("W")) { 
						sourceFrom = "("+(thisRow)+","+(thisCol+1)+")";

					}

					// Construct the label that will be used to retrieve this node
					// from the map of nodes.
					String sourceLabel = sourceFrom+" "+approachDir;

					// If source node already exists,
					// get it from the map.
					// If this node does not exist yet, 
					// construct it and put it in the map.
					Node sourceNode;
					if(nodesMap.containsKey(sourceLabel)) {
						sourceNode = nodesMap.get(sourceLabel);
					} else {
						sourceNode = new Node(sourceLabel, sourceFrom, sourceTo);
						nodesMap.put(sourceLabel,sourceNode);
					}



					// DestinationLabel:
					//
					// Now we need to find the destination node.
					// We are potentially constructing this node too.
					//
					// Specify the intersections it connects from and to.
					// Destination node connects from the midpoint to the destination point.
					String destFrom = "("+thisRow+","+thisCol+")";
					String destTo = "";



					// Loop over each of the valid turns specified in the input file
					// if we are approaching the intersection (thisRow, thisCol) 
					// from the given direction, thisDir. 
					int nturns = chunk.length() - 1;
					String turn = "";
					for(int k=1; k<chunk.length(); k++) { 

						// Construct the label of the node 
						// that this allowed turn leads to: 
						String connectLabel = "";

						// The valid turn that is specified
						turn = chunk.substring(k,k+1);

						// Direction of next node in sequence:
						String destDir = "";
						if(approachDir.equals("N")) {
							if(turn.equals("L")) { 
								destDir = "W";
								destTo = "("+(thisRow)+","+(thisCol-1)+")";

							} else if(turn.equals("R")) { 
								destDir = "E";
								destTo = "("+(thisRow)+","+(thisCol+1)+")";

							} else if(turn.equals("F")) { 
								destDir = approachDir;
								destTo = "("+(thisRow-1)+","+(thisCol)+")";
							}

						} else if(approachDir.equals("S")) {
							if(turn.equals("L")) { 
								destDir = "E";
								destTo = "("+(thisRow)+","+(thisCol+1)+")";

							} else if(turn.equals("R")) { 
								destDir = "W";
								destTo = "("+(thisRow)+","+(thisCol-1)+")";

							} else if(turn.equals("F")) { 
								destDir = approachDir;
								destTo = "("+(thisRow+1)+","+(thisCol)+")";

							}

						} else if(approachDir.equals("E")) {
							if(turn.equals("L")) { 
								destDir = "N";
								destTo = "("+(thisRow-1)+","+(thisCol)+")";

							} else if(turn.equals("R")) { 
								destDir = "S";
								destTo = "("+(thisRow+1)+","+(thisCol)+")";

							} else if(turn.equals("F")) { 
								destDir = approachDir;
								destTo = "("+(thisRow+1)+","+(thisCol+1)+")";

							}

						} else if(approachDir.equals("W")) {
							if(turn.equals("L")) { 
								destDir = "S";
								destTo = "("+(thisRow+1)+","+(thisCol)+")";

							} else if(turn.equals("R")) { 
								destDir = "N";
								destTo = "("+(thisRow-1)+","+(thisCol)+")";

							} else if(turn.equals("F")) { 
								destDir = approachDir;
								destTo = "("+(thisRow)+","+(thisCol-1)+")";

							}

						}

						String destLabel = destFrom + " " + destDir;

						Node destNode;
						if(nodesMap.containsKey(destLabel)) {
							destNode = nodesMap.get(destLabel);
						} else {
							destNode = new Node(destLabel, destFrom, destTo);
							nodesMap.put(destLabel, destNode);
						}

						sourceNode.link(destNode);

					}

				} // end process each turn on this line

			} // end for each line in input file

			i++;
		}

		for(Node n : nodesMap.values()) {
			n.toString();
			n.printLinks();
		}
	}

	/** Iterate through each line of this maze's inputs 
	  * to determine maximum row and column (maze dimensions). */
	public Dimension getMaxDimensions() {
		List<Integer> rows = new ArrayList<Integer>();
		List<Integer> cols = new ArrayList<Integer>();

		// Iterate through all lines and extract row/column indexes.
		// These will be used to find max row and max col,
		// which will (elsewhere) be used to build the graph.
		Scanner lineScan; 
		int i=0;
		for(String thisLine : this.lines) {
			// Treat first line differently
			if(i==0) { 
				// First line contains start and end locations.
				// Specifically:
				// <StartRow> <StartCol>  (StartDir) <EndRow> <EndCol> 
				String[] tokens = thisLine.split(" ");
				rows.add(Integer.parseInt(tokens[0]));
				cols.add(Integer.parseInt(tokens[1]));
				rows.add(Integer.parseInt(tokens[3]));
				cols.add(Integer.parseInt(tokens[4]));

			} else {
				// Remaining lines contain one coordinate each.
				String[] tokens = thisLine.split(" ");
				rows.add(Integer.parseInt(tokens[0]));
				cols.add(Integer.parseInt(tokens[1]));
			}

			i++;
		}
		
		int maxRow = getMax(rows);
		int maxCol = getMax(cols);
		this.dim = new Dimension(maxRow,maxCol);
		return this.dim;
	}

	/** Get the maximum number in a list of integers. */
	private int getMax(List<Integer> list) {
		int max = 0;
		for(Integer i : list) {
			max = Math.max(max,i);
		}
		return max;
	}
}

class Dimension {
	int rows, cols;
	public Dimension(int rows, int cols) { 
		this.rows = rows; this.cols = cols;
	}
}

class Node { 
	public String name, from, to; // Name of the intersections 
							// this Node connects from and to
	public List<Node> siblings; // List of all Nodes that are siblings 
								// (connected to) this node
	public boolean isHead, isTail;  // head is maze entrance,
									// tail is maze exit 

	public Node(String name, String from, String to){
		this.name = name;
		this.from = from;
		this.to = to;
		this.siblings = new ArrayList<Node>();
		this.isHead = false;
		this.isTail = false;
	}
	public void link(Node n) { 
		siblings.add(n);
	}
	public void printLinks() { 
		System.out.print("Node "+name+" list of siblings: ");
		System.out.println(this.siblings);
		System.out.println();
	}
	public void setHead() {
		this.isHead = true;
	}
	public void setTail() { 
		this.isTail = true;
	}
	public String toString() {
		return "Node "+name;
		//return "\"Node: " + this.from + "--->" + this.to + "\"";
		//return this.from+"--->"+this.to;
	}
}

